//
// Author:  Christoph Dalitz
// Version: 1.0 from 2013/01/03
//
// Computes Reisfelds Symmetry-Transformation
// and saves it as "reisfeld.png"
// Does *not* follow the conference paper from 1990,
// but the IJCV paper from 1995
//

#include <stdio.h>
#include <math.h>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include "symmetry_transform.h"

//int radius_cutoff = 5;  // actual cutoff at radius_cutoff * opt_s

// command line arguments
char* opt_infile = NULL;
char* opt_outfile = NULL;
double opt_s = 0.0;
int opt_r = 0;
bool opt_log = false;
bool opt_trace = false;
const char* usage = "Usage:\n\tsymmetry -r <r> -pos <infile>\nOptions:\n"
  "\t-s <s> value for sigma\n"
  "\t-r <r> value for r (cutoff radius)\n"
  "\t-gradlog\n"
  "\t       transform gradient strength as log(1+|G|)\n"
  "\t-o <outpng>\n"
  "\t       draw detected symmetry as rectangle in file <outpng>\n"
  "\t-trace write trace of interim images with prefix 'trace-'\n"
  "\t-v     print version and exit\n";


int main(int argc, char** argv)
{
  cv::Mat img, grey, gradx, grady, gradabs, gradangle, Dvals, Alphavals, result, Cij, Phiij;
  int r,i,x,y,dx,dy,max_r,max_radius;
  double minval,maxval,symmetry,gx,gy,normfac;
  cv::Point minloc,maxloc;

  // parse command line
  for (i=1; i<argc; i++) {
    if (0==strcmp("-s", argv[i])) {
      i++; opt_s = atof(argv[i]);
    }
    if (0==strcmp("-r", argv[i])) {
      //i++; opt_s = (int)(atoi(argv[i])/3.0 + 0.5);
      i++; opt_r = atoi(argv[i]);
    }
    else if (0==strcmp("-gradlog", argv[i])) {
      opt_log = true;
    }
    else if (0==strcmp("-o", argv[i])) {
      i++; opt_outfile = argv[i];
    }
    else if (0==strcmp("-v", argv[i])) {
      printf("%s\n", VERSION);
      return 0;
    }
    else if (0==strcmp("-trace", argv[i])) {
      opt_trace = true;
    }
    else if (argv[i][0] == '-') {
      fprintf(stderr, usage);
      return 1;
    }
    else {
      opt_infile = (char*)argv[i];
    } 
  }
  if (opt_s == 0.0 || opt_r == 0 || !opt_infile) {
    fprintf(stderr, usage);
    return 1;
  }

  img = cv::imread(opt_infile);
  if (!img.data) {
    fprintf(stderr, "cannot open '%s'\n", opt_infile);
    return 1;
  }
  cv::cvtColor(img,grey,CV_RGB2GRAY);

  // compute gradient images
  cv::Sobel(grey, gradx, CV_64F, 1, 0);
  cv::Sobel(grey, grady, CV_64F, 0, 1);
  gradabs.create(grey.rows, grey.cols, CV_64F);
  gradangle.create(grey.rows, grey.cols, CV_64F);
  for (y=0; y<grey.rows; y++) {
    for (x=0; x<grey.cols; x++) {
      gx = gradx.at<double>(y,x);
      gy = grady.at<double>(y,x);
      if (opt_log) {
        gradabs.at<double>(y,x) = log(1+sqrt(gx*gx+gy*gy));
      } else {
        gradabs.at<double>(y,x) = sqrt(gx*gx+gy*gy);
      }
      gradangle.at<double>(y,x) = atan2(gy,gx);
    }
  }

  // precompute some values
  //max_radius = (int)radius_cutoff*opt_s;
  max_radius = opt_r;
  Dvals.create(2*max_radius+1,2*max_radius+1,CV_64F);
  normfac = 1.0/(sqrt(2*M_PI*opt_s));
  for (y=0; y<Dvals.rows; y++) {
    for (x=0; x<Dvals.cols; x++) {
      dx = x - max_radius;
      dy = y - max_radius;
      Dvals.at<double>(y,x) = normfac*exp(-sqrt(dx*dx+dy*dy)/opt_s);
    }
  }
  Alphavals.create(2*max_radius+1,2*max_radius+1,CV_64F);
  for (y=0; y<Alphavals.rows; y++) {
    for (x=0; x<Alphavals.cols; x++) {
      dx = x - max_radius;
      dy = y - max_radius;
      Alphavals.at<double>(y,x) = atan2(dy,dx);
    }
  }
  
  //
  // compute symmetry transform
  //
  result.create(grey.rows, grey.cols, CV_64F);
  double maxC,Phi,C,thetai,thetaj,sinval;
  Cij.create(2*max_radius+1,2*max_radius+1,CV_64F);
  Phiij.create(2*max_radius+1,2*max_radius+1,CV_64F);
  for (y=0; y<grey.rows; y++) {
    for (x=0; x<grey.cols; x++) {
      // adjust radius so that it does not extend beyond the image border
      max_r = max_radius;
      if (y < max_r)
        max_r = y;
      if (y >= grey.rows - max_r)
        max_r = grey.rows - y - 1;
      if (x < max_r)
        max_r = x;
      if (x >= grey.cols - max_r)
        max_r = grey.cols - x - 1;
      if (max_r < 1) {
        result.at<double>(y,x) = 0.0;
        continue;
      }

      maxC = Phi = 0.0;
      for (r=1; r<=max_r; r++) {
        dy = r;
        for (dx=-r; dx<=r; dx++) {
          thetai = gradangle.at<double>(y+dy,x+dx);
          thetaj = gradangle.at<double>(y-dy,x-dx);
          C = Dvals.at<double>(max_radius+dy,max_radius+dx)*
            (1-cos(thetai-thetaj))*
            (1-cos(thetai+thetaj-2*Alphavals.at<double>(max_radius+dy,max_radius+dx)))*
            gradabs.at<double>(y+dy,x+dx)*gradabs.at<double>(y-dy,x-dx);
          /*
          C = normfac*exp(-sqrt(dy*dy+dx*dx)/opt_s)*
            (1-cos(thetai-thetaj))*
            (1-cos(thetai+thetaj-2*atan2(dy,dx)))*
            gradabs.at<double>(y+dy,x+dx)*gradabs.at<double>(y-dy,x-dx);
          */
          Cij.at<double>(max_radius+dy,max_radius+dx) = C;
          Phiij.at<double>(max_radius+dy,max_radius+dx) = (thetai+thetaj)/2;
          if (C > maxC) {
            maxC = C;
            Phi = (thetai+thetaj)/2;
          }
        }
        dx = r;
        for (dy=-r+1; dy<r; dy++) {
          thetai = gradangle.at<double>(y+dy,x+dx);
          thetaj = gradangle.at<double>(y-dy,x-dx);
          C = Dvals.at<double>(max_radius+dy,max_radius+dx)*
            (1-cos(thetai-thetaj))*
            (1-cos(thetai+thetaj-2*Alphavals.at<double>(max_radius+dy,max_radius+dx)))*
            gradabs.at<double>(y+dy,x+dx)*gradabs.at<double>(y-dy,x-dx);
          /*
          C = normfac*exp(-sqrt(dy*dy+dx*dx)/opt_s)*
            (1-cos(thetai-thetaj))*
            (1-cos(thetai+thetaj-2*atan2(dy,dx)))*
            gradabs.at<double>(y+dy,x+dx)*gradabs.at<double>(y-dy,x-dx);
          */
          Cij.at<double>(max_radius+dy,max_radius+dx) = C;
          Phiij.at<double>(max_radius+dy,max_radius+dx) = (thetai+thetaj)/2;
          if (C > maxC) {
            maxC = C;
            Phi = (thetai+thetaj)/2;
          }
        }
      }

      // collect values relative to maximum direction Phi
      symmetry = 0.0;
      for (r=1; r<=max_r; r++) {
        dy = r;
        for (dx=-r; dx<=r; dx++) {
          sinval = sin(Phiij.at<double>(max_radius+dy,max_radius+dx)-Phi);
          symmetry += Cij.at<double>(max_radius+dy,max_radius+dx)*sinval*sinval;
          //symmetry += Cij.at<double>(max_radius+dy,max_radius+dx);
        }
        dx = r;
        for (dy=-r+1; dy<r; dy++) {
          sinval = sin(Phiij.at<double>(max_radius+dy,max_radius+dx)-Phi);
          symmetry += Cij.at<double>(max_radius+dy,max_radius+dx)*sinval*sinval;
          //symmetry += Cij.at<double>(max_radius+dy,max_radius+dx);
        }
      }
      result.at<double>(y,x) = symmetry;
    }
  }


  // draw symmetry weight image
  if (opt_trace) {
    float2grey(result,grey);
    cv::imwrite("trace-reisfeld95.png", grey);
  }

  // draw maximum symmetry point in original image
  cv::minMaxLoc(result, &minval, &maxval, &minloc, &maxloc);
  r = (int)3*opt_s;
  printf("point=(%i,%i)\n", maxloc.x, maxloc.y);
  if (opt_outfile) {
    cv::rectangle(img,cv::Point(maxloc.x-r,maxloc.y-r),cv::Point(maxloc.x+r,maxloc.y+r),cv::Scalar(0,0,255));
    cv::imwrite(opt_outfile, img);
  }

  return 0;
}
