//
// Author:  Christoph Dalitz
// Version: 1.0 from 2013/01/03
//
// Computes the symmetry transformation after
// VISAPP 2013 paper by Dalitz, Pohle-Froehlich, and Bolten
// and return the maximum rotational symmetry point
//

#include <stdio.h>
#include <math.h>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include "symmetry_transform.h"

#define TRACEPREFIX "trace-"

// command line arguments
char* opt_infile = NULL;
int opt_r = 0;
bool opt_log = false;
bool opt_norm = false;
bool opt_trace = false;
char* opt_outfile = NULL;
const char* usage = "Usage:\n\tsymmetry -r <r> [options] <infile>\nOptions:\n"
  "\t-r <r> maximum radius of symmetry transform\n"
  "\t-o <outpng>\n"
  "\t       draw detected symmetry as rectangle in file <outpng>\n"
  "\t-gradlog\n"
  "\t       transform gradient strength as log(1+|G|)\n"
  "\t-trace write trace of interim images with prefix '" TRACEPREFIX "'\n"
  "\t-v     print version and exit\n";
//  "\t-norm   normalize result by mean of gradient square in region\n"


int main(int argc, char** argv)
{
  cv::Mat img, grey, bgr, result, result_r, gradx, grady;
  int r,i;
  cv::Point maxpoint(0,0);

  // parse command line
  for (i=1; i<argc; i++) {
    if (0==strcmp("-r", argv[i])) {
      i++; opt_r = atoi(argv[i]);
    }
    else if (0==strcmp("-o", argv[i])) {
      i++; opt_outfile = argv[i];
    }
    else if (0==strcmp("-v", argv[i])) {
      printf("%s\n", VERSION);
      return 0;
    }
    else if (0==strcmp("-gradlog", argv[i])) {
      opt_log = true;
    }
    else if (0==strcmp("-trace", argv[i])) {
      opt_trace = true;
    }
    else if (0==strcmp("-norm", argv[i])) {
      opt_norm = true;
      printf("Warning: option '-norm' currently ignored\n");
      fflush(stdout);
    }
    else if (argv[i][0] == '-') {
      fprintf(stderr, usage);
      return 1;
    }
    else {
      opt_infile = (char*)argv[i];
    } 
  }
  if (!opt_r || !opt_infile) {
    fprintf(stderr, usage);
    return 1;
  }

  // load image and call symmetry transform
  img = cv::imread(opt_infile);
  if (!img.data) {
    fprintf(stderr, "cannot open '%s'\n", opt_infile);
    return 1;
  }
  cv::cvtColor(img,grey,CV_RGB2GRAY);
  if (opt_log) {
    GradientNormalization gradnorm;
    symmetry_cd(grey, result, result_r, opt_r, &gradnorm);
  } else {
    symmetry_cd(grey, result, result_r, opt_r);
  }
  if (opt_trace) {
    float2grey(result,grey);
    cv::imwrite(TRACEPREFIX "symmetry.png", grey);
  }

  // find all local maxima and sort them by symmetry score
  // BEWARE: sorting is in ascending order!
  SymmetryPointVector sp;
  if (0 >= points_local_maxima(result, sp)) {
    fprintf(stderr, "no symmetry points found\n");
    return -2;
  }
  std::sort(sp.begin(), sp.end());

  // find first candidate point with edge directedness < 0.27
  cv::Sobel(result, gradx, CV_64F, 1, 0);
  cv::Sobel(result, grady, CV_64F, 0, 1);
  if (opt_trace) {
    float2bgr(result,bgr);
  }
  for (i=(int)sp.size()-1; i>=0; i--) {
    if (edge_directedness_cd(gradx, grady, sp[i].point, 3, 16) < 0.27) {
      maxpoint = sp[i].point;
      if (opt_trace) {
        bgr.at<cv::Vec3b>(sp[i].point) = cv::Vec3b(0,255,0);
        r = (int)result_r.at<double>(sp[i].point);
        cv::rectangle(bgr,cv::Point(sp[i].point.x-r,sp[i].point.y-r),cv::Point(sp[i].point.x+r,sp[i].point.y+r),cv::Scalar(0,255,0));
      }
      break;
    } else {
      if (opt_trace)
        bgr.at<cv::Vec3b>(sp[i].point) = cv::Vec3b(0,0,255);
        r = (int)result_r.at<double>(sp[i].point);
        cv::rectangle(bgr,cv::Point(sp[i].point.x-r,sp[i].point.y-r),cv::Point(sp[i].point.x+r,sp[i].point.y+r),cv::Scalar(0,0,255));
    }
  }
  if (opt_trace) {
    cv::imwrite(TRACEPREFIX "localmaxpoints.png", bgr);
  }

  // draw maximum symmetry point in original image
  if (maxpoint.x == 0 && maxpoint.y == 0) {
    fprintf(stderr, "no symmetry point found\n");
    return 3;
  }
  r = (int)result_r.at<double>(maxpoint);
  printf("point=(%i,%i) radius=%i\n", maxpoint.x, maxpoint.y, r);
  if (opt_outfile) {
    cv::rectangle(img,cv::Point(maxpoint.x-r,maxpoint.y-r),cv::Point(maxpoint.x+r,maxpoint.y+r),cv::Scalar(0,0,255));
    cv::imwrite(opt_outfile, img);
  }

  return 0;
}
