//
// Author:  Christoph Dalitz, Regina Pohle-Froehlich
// Version: 1.0 from 2013/01/03
//

#include <math.h>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include "symmetry_point.h"
#include <string>
#include <stdio.h>
#include <stdlib.h>

//-----------------------------------------------------------------
// Convert float image to greyscale image.
// Arguments:
//  fimg = float image (input)
//  gimg = greyscale image (output)
//-----------------------------------------------------------------
void float2grey(const cv::Mat &fimg, cv::Mat &gimg) {
  double minval,maxval;
  cv::minMaxLoc(fimg, &minval, &maxval);
  cv::convertScaleAbs(fimg, gimg, 255.0/(maxval-minval), -minval*255.0/(maxval-minval));
}

//-----------------------------------------------------------------
// Convert float image to BGR image.
// Arguments:
//  fimg = float image (input)
//  cimg = color image (output)
//-----------------------------------------------------------------
void float2bgr(const cv::Mat &fimg, cv::Mat &cimg) {
  cv::Mat gimg;
  float2grey(fimg, gimg);
  cv::cvtColor(gimg, cimg, CV_GRAY2BGR);
}

//-----------------------------------------------------------------
// Gradient normalization function
// Derive your custom normalization routine for special normalizations
// The call operator has two arguments:
//  gx = x-component of gradient (input and output)
//  gy = y-component of gradient (input and output)
//-----------------------------------------------------------------
struct GradientNormalization {
  //GradientNormalization(void) {}
  void operator()(double &gx, double &gy) const {
    // logarithmic normalization
    double norm = sqrt(gx*gx + gy*gy);
    if (norm > 0.1) {
      double lognorm = log(1+norm);
      gx = gx*lognorm/norm;
      gy = gy*lognorm/norm;
    }
  }
};

//-----------------------------------------------------------------
// Compute symmetry transform from a greyscale image.
// To each point, a symmetry score value and a radius is assigned
// Arguments:
//  grey     = greyscale image (input)
//  result   = symmetry scores as CV_64F image (output)
//  result_r = symmetry radius as CV_64F image (output)
//  max_r    = maximum radius examined (input)
//  gradnorm = optional gradient normalization function (default = NULL)
//-----------------------------------------------------------------
void symmetry_cd(const cv::Mat &grey, cv::Mat &result, cv::Mat &result_r, int radius, GradientNormalization *gradnorm/*=NULL*/)
{
  double symmetry, scalarprod;
  int x,y,dx,dy,r,max_r,max_radius;
  double maxval,gx,gy;
  cv::Mat gradx, grady;

  cv::Sobel(grey, gradx, CV_64F, 1, 0);
  cv::Sobel(grey, grady, CV_64F, 0, 1);
  result.create(grey.rows, grey.cols, CV_64F);
  result_r.create(grey.rows, grey.cols, CV_64F);

  /***************************************
  // save grad image
  cv::Mat gradabs, gradabsgrey;
  gradabs.create(grey.rows, grey.cols, CV_64F);
  gradabsgrey.create(grey.rows, grey.cols, CV_8U);
  for (y=0; y<grey.rows; y++) {
    for (x=0; x<grey.cols; x++) {
      gx = gradx.at<double>(y,x);
      gy = grady.at<double>(y,x);
      gradabs.at<double>(y,x) = sqrt(gx*gx+gy*gy);
    }
  }
  float2grey(gradabs,gradabsgrey);
  cv::imwrite("gradabs.png", gradabsgrey);
  ****************************************/


  // optionally transform gradient strength
  if (gradnorm) {
    for (y=0; y<grey.rows; y++) {
      for (x=0; x<grey.cols; x++) {
        if (gradnorm) {
          gx = gradx.at<double>(y,x);
          gy = grady.at<double>(y,x);
          (*gradnorm)(gx,gy);
          gradx.at<double>(y,x) = gx;
          grady.at<double>(y,x) = gy;
        }
      }
    }   
  }

  // compute symmetry weight
  std::vector<double> weight_at_r(radius+1);
  weight_at_r[0] = 0.0; // not needed
  for (y=0; y<grey.rows; y++) {
    for (x=0; x<grey.cols; x++) {
      // adjust radius so that it does not extend beyond the image border
      max_radius = radius;
      if (y < max_radius)
        max_radius = y;
      if (y >= grey.rows - max_radius)
        max_radius = grey.rows - y - 1;
      if (x < max_radius)
        max_radius = x;
      if (x >= grey.cols - max_radius)
        max_radius = grey.cols - x - 1;

      // compute weight in window with radius max_radius
      if (max_radius < 1) {
        result.at<double>(y,x) = 0.0;
        result_r.at<double>(y,x) = 0.0;
      } else {
        symmetry = 0.0;
        for (r=1; r<=max_radius; r++) {
          dy = r;
          for (dx=-r; dx<=r; dx++) {
              scalarprod = gradx.at<double>(y+dy,x+dx)*gradx.at<double>(y-dy,x-dx)
                + grady.at<double>(y+dy,x+dx)*grady.at<double>(y-dy,x-dx);
              symmetry += -scalarprod;
          }
          dx = r;
          for (dy=-r+1; dy<r; dy++) {
              scalarprod = gradx.at<double>(y+dy,x+dx)*gradx.at<double>(y-dy,x-dx)
                + grady.at<double>(y+dy,x+dx)*grady.at<double>(y-dy,x-dx);
              symmetry += -scalarprod;
          }
          weight_at_r[r] = symmetry;
        }
        // find radius of maximum symmetry
        max_r = 1;
        maxval = weight_at_r[1];
        for (r=2; r<=max_radius; r++) {
          if (weight_at_r[r] > maxval) {
            max_r = r;
            maxval = weight_at_r[r];
          }
        }
        result.at<double>(y,x) = maxval;
        result_r.at<double>(y,x) = (double)max_r;
      }
    }
  }

  return;
}

//-----------------------------------------------------------------
// Find candidate symmetry centers in symmetry image as local
// maxima of the symmetry score. Returns the number of points found.
// Arguments:
//  symmetry        = symmetry transform image filename (input)
//  candidatepoints = symmetry candidate points (output)
//  windowsize      = window size for local maximum search (input)
//-----------------------------------------------------------------
int points_local_maxima(const cv::Mat &symmetry, SymmetryPointVector &candidatepoints, int windowsize /*=3*/)
{
	if (windowsize%2 == 0)
	{
		fprintf(stderr, "odd number for windowsize parameter expected\n");
		return -1;
	}
	candidatepoints.clear();
	double minv, maxv;
	cv::minMaxLoc(symmetry, &minv, &maxv);
	int s = windowsize/2;
	int maxindex;
	int maxi=s*windowsize+s;
	for (int y=s; y<=symmetry.rows-s-1; y++)
		for (int x=s; x<= symmetry.cols-s-1; x++)
		{
			double maxivalue=minv;			
			for (int yy=0; yy<=2*s; yy++)
				for (int xx=0; xx<=2*s; xx++)
				{
					double d = symmetry.at<double>(y+yy-s,x+xx-s);
					if (d > maxivalue)
					{
						maxindex=yy*windowsize+xx;
						maxivalue=d;
					}
				}
			if (maxindex == maxi)
			{
				candidatepoints.push_back(SymmetryPoint(cv::Point(x,y),symmetry.at<double>(y,x)));
			}
		}

	return candidatepoints.size();
}


//-----------------------------------------------------------------
// Compute the edge dispersion at the given point.
// Arguments:
//  gradx    = x-component of gradient image (input)
//  grady    = y-component of gradient image (input)
//  point    = the point to be examined (input)
//  radius   = the radius of the window to be examined (input)
//  cutoff   = ignore an image border of width cutoff (input)
//-----------------------------------------------------------------
double edge_dispersion_cd(const cv::Mat &gradx, const cv::Mat &grady, const cv::Point &point, int radius /* = 3*/, int cutoff /* = 0*/)
{
  //  when G_i are the gradient vectors inside the window,
  //  then we compute
  //
  //    1 - R / (n max{ ||G_i|| })  with  R = || \sum_i G_i ||
  //
  //  The problem is that we do not know how to mirror the vectors
  //  G_i (if we do not do this, the dispersion is always maximal).
  //  We therefore first mirror negative y-values, and then negative
  //  x-values, and eventaully choose the smaller dispersion of both.

  int x,y,left,right,top,bot,n;
  double G2, maxG2; // square of absolute value
  double gx, gy;    // gradient values
  double dispersion1, dispersion2;
  double dispersion1_x = 0.0;
  double dispersion1_y = 0.0;
  double dispersion2_x = 0.0;
  double dispersion2_y = 0.0;

  left  = std::max(cutoff, point.x - radius);
  right = std::min(gradx.cols - cutoff - 1, point.x + radius);
  top   = std::max(cutoff, point.y - radius);
  bot   = std::min(gradx.rows - cutoff - 1, point.y + radius);

  maxG2 = 0.0; n=0;
  for (y = top; y <= bot; y++) {
    for (x = left; x <= right; x++) {
      n++;
      gx = gradx.at<double>(y,x); gy = grady.at<double>(y,x);
      G2 = gx*gx + gy*gy;
      if (G2 > maxG2)
        maxG2 = G2;
      if (gx < 0) {
        dispersion1_x -= gx; dispersion1_y -= gy;
      } else {
        dispersion1_x += gx; dispersion1_y += gy;        
      }
      if (gy < 0) {
        dispersion2_x -= gx; dispersion2_y -= gy;
      } else {
        dispersion2_x += gx; dispersion2_y += gy;        
      }
    }
  }

  // apply formula for dispersion
  if (maxG2 > 0.0) {
    dispersion1 = 1 - sqrt((dispersion1_x*dispersion1_x + dispersion1_y*dispersion1_y) / maxG2)/n;
    dispersion2 = 1 - sqrt((dispersion2_x*dispersion2_x + dispersion2_y*dispersion2_y) / maxG2)/n;
  } else {
    return 0.0;
  }

  // choose the smaller dispersion
  return std::min(dispersion1, dispersion2);
}


//-----------------------------------------------------------------
// Compute the edge directedness as the frequency of the most frequent
// gradient direction.
// Arguments:
//  gradx    = x-component of gradient image (input)
//  grady    = y-component of gradient image (input)
//  point    = the point to be examined (input)
//  radius   = the radius of the window to be examined (input)
//  cutoff   = ignore an image border of width cutoff (input)
//-----------------------------------------------------------------
double edge_directedness_cd(const cv::Mat &gradx, const cv::Mat &grady, const cv::Point &point, int radius /*= 3*/, int nbins /*= 16*/, int cutoff /*= 0*/)
{
  int x,y,left,right,top,bot,i;
  double G, sumG;  // absolute values of gradient
  double gx, gy;   // gradient values
  double angle, maxhist, pi_n;
  double* hist = new double[nbins];
  
  pi_n = M_PI / nbins;
  left  = std::max(cutoff, point.x - radius);
  right = std::min(gradx.cols - cutoff - 1, point.x + radius);
  top   = std::max(cutoff, point.y - radius);
  bot   = std::min(gradx.rows - cutoff - 1, point.y + radius);

  sumG = 0.0;
  for (i=0; i<nbins; i++) hist[i] = 0.0;
  for (y = top; y <= bot; y++) {
    for (x = left; x <= right; x++) {
      gx = gradx.at<double>(y,x); gy = grady.at<double>(y,x);
      G = sqrt(gx*gx + gy*gy);
      sumG += G;
      angle = atan2(gy,gx);
      // shift angle bins by pi/n so that zero etc. lie in middle of bin
      if (angle < -pi_n) angle += 2*M_PI;
      i = (int)((angle+pi_n)*nbins/(2*M_PI));
      hist[i] += G;
      //hist[(i+4)%8] += G;
    }
  }
  maxhist = 0.0;
  for (i=0; i<nbins; i++) 
    if (hist[i] > maxhist) maxhist = hist[i];

  delete[] hist;

  if (sumG == 0.0)
    return 0.0;
  else
    return (maxhist/sumG);
  //return (maxhist/(2*sumG));
}

//-----------------------------------------------------------------
// Hu moments of region around point.
// Arguments:
//  image     = complete image (input)
//  point     = the point to be examined (input)
//  radius    = the radius of the window to be examined (input)
//  humoments = the computed moments (output, must be preallocated of size 8)
//-----------------------------------------------------------------
void humoments_cd(double* humoments, const cv::Mat &image, const cv::Point &point, int radius /*= 3*/, int cutoff /*= 0*/)
{
  int left,right,top,bot;
  cv::Mat roi, roigrey;
  cv::Moments m;

  left  = std::max(cutoff, point.x - radius);
  right = std::min(image.cols - cutoff - 1, point.x + radius);
  top   = std::max(cutoff, point.y - radius);
  bot   = std::min(image.rows - cutoff - 1, point.y + radius);

  roi = image(cv::Rect(left,top,right-left+1,bot-top+1));
  float2grey(roi, roigrey); // normalize values in region
  m = cv::moments(roigrey);
  cv::HuMoments(m, humoments);
}

//-----------------------------------------------------------------
// Bamieh moment invariants (BMI) of region around point.
// Arguments:
//  image     = complete image (input)
//  point     = the point to be examined (input)
//  radius    = the radius of the window to be examined (input)
//  bmi       = the computed moments (output, must be preallocated of size 2)
//-----------------------------------------------------------------
void bmimoments_cd(double* bmi, const cv::Mat &image, const cv::Point &point, int radius /*= 3*/, int cutoff /*= 0*/)
{
  int left,right,top,bot;
  cv::Mat roi, roigrey;
  cv::Moments m;

  left  = std::max(cutoff, point.x - radius);
  right = std::min(image.cols - cutoff - 1, point.x + radius);
  top   = std::max(cutoff, point.y - radius);
  bot   = std::min(image.rows - cutoff - 1, point.y + radius);

  roi = image(cv::Rect(left,top,right-left+1,bot-top+1));
  float2grey(roi, roigrey); // normalize values in region
  m = cv::moments(roigrey);
  bmi[0] = m.nu02*m.nu20 - m.nu11*m.nu11;
  bmi[1] = (m.nu03*m.nu30 - m.nu21*m.nu12)*(m.nu03*m.nu30 - m.nu21*m.nu12)
    - 4*(m.nu03*m.nu12 - m.nu21*m.nu21)*(m.nu30*m.nu21 - m.nu12*m.nu12);
}

//-----------------------------------------------------------------
// Compute the ratio of the the two eigenvalues of teh covariance matrix
// of the window with the given radius around point.
// Arguments:
//  image     = complete image (input)
//  point     = the point to be examined (input)
//  radius    = the radius of the window to be examined (input)
//  cutoff   = ignore an image border of width cutoff (input)
//-----------------------------------------------------------------
double coveigenratio_cd(const cv::Mat &image, const cv::Point &point, int radius /*= 3*/, int cutoff /*= 0*/)
{
  int x,y,left,right,top,bot,i,j;
  cv::Mat cov(cv::Size(2,2), CV_64F);
  cv::Mat eigenvalues;
  double sumf;  // normalization factor
  double value,dy,dx;

  left  = std::max(cutoff, point.x - radius);
  right = std::min(image.cols - cutoff - 1, point.x + radius);
  top   = std::max(cutoff, point.y - radius);
  bot   = std::min(image.rows - cutoff - 1, point.y + radius);

  // compute covariance matrix
  for (i=0; i<2; i++)
    for (j=0; j<2; j++)
      cov.at<double>(i,j) = 0.0;
  sumf = 0.0;
  for (y = top; y <= bot; y++) {
    for (x = left; x <= right; x++) {
      if (image.type() == CV_64F) {
        value = image.at<double>(y,x);
      }
      else if (image.type() == CV_32F) {
        value = (double)image.at<float>(y,x);
      }
      else if (image.type() == CV_8U) {
        value = (double)image.at<uchar>(y,x);
      }
      else if (image.type() == CV_16U) {
        value = (double)image.at<ushort>(y,x);
      }
      sumf += value;
      dx = x - point.x; dy = y - point.y;
      cov.at<double>(0,0) += value*dx*dx;
      cov.at<double>(0,1) += value*dx*dy;
      cov.at<double>(1,0) += value*dy*dx;
      cov.at<double>(1,1) += value*dy*dy;
    }
  }
  for (i=0; i<2; i++)
    for (j=0; j<2; j++)
      cov.at<double>(i,j) /= sumf;

  // determine eigenvalue ratio
  cv::eigen(cov, eigenvalues);

  cv::Size s = eigenvalues.size();
  //printf("Size of eigenvalues = %ix%i\n", s.width, s.height);
  //printf("Eigenvalues = (%f,%f)", eigenvalues.at<double>(0,0), eigenvalues.at<double>(1,0));

  if (eigenvalues.at<double>(0,0) == 0.0 && eigenvalues.at<double>(1,0) == 0.0)
    return 0.0;
  if (eigenvalues.at<double>(0,0) > eigenvalues.at<double>(1,0))
    return (eigenvalues.at<double>(1,0) / eigenvalues.at<double>(0,0));
  else
    return (eigenvalues.at<double>(0,0) / eigenvalues.at<double>(1,0));
}


//-----------------------------------------------------------------
// Returns the number of antiparallel gradient bins that have a
// frequency count larger than count_threshold
// inside the window with the given radius around point.
// Arguments:
//  gradx    = x-component of gradient image (input)
//  grady    = y-component of gradient image (input)
//  point    = the point to be examined (input)
//  radius   = the radius of the window to be examined (input)
//  max_frequency = frequency count of largest bin (output)
//  weighted = whether edges shall be weighted by gradient strength (input)
//  count_threshold = threshold when a bin is considered "occupied" (input)
//  nbins    = number of bins in direction histogram (input)
//-----------------------------------------------------------------
double antiparallelity_rpf(const cv::Mat &gradx, const cv::Mat &grady, const cv::Point &point, int radius, double &max_frequency, bool weighted /*= false*/, double count_threshold /*= 0.1*/, int nbins /*= 8*/)
{
  double scalarprod, result;
  int x,y,dx,dy,r,i,count;
  double v1,v2;  // norms of the opposite gradients
  double angle;  // angle between gradients
  double sumG;   // normalization factor (sum of all gradient norms)
  // threshold when angle between two vectors can be considered antiparallel
  double parallelity_threshold = -0.975;
  // pi divided by number of bins (shift for histogram bins)
  double pi_n = M_PI / nbins;

  // compute antiparallelity
  std::vector<double> bin(nbins);
  for (i=0; i<nbins; i++) bin[i] = 0.0;
  sumG = 0.0;
  x = point.x;
  y = point.y;
  for (r=1; r<=radius; r++) {
    dy = r;
    for (dx=-r; dx<=r; dx++) {
      scalarprod = gradx.at<double>(y+dy,x+dx)*gradx.at<double>(y-dy,x-dx)
        + grady.at<double>(y+dy,x+dx)*grady.at<double>(y-dy,x-dx);
      if (scalarprod < 0.0) { // only when in opposite directions
        v1 = sqrt(gradx.at<double>(y+dy,x+dx)*gradx.at<double>(y+dy,x+dx) +
                  grady.at<double>(y+dy,x+dx)*grady.at<double>(y+dy,x+dx));
        v2 = sqrt(gradx.at<double>(y-dy,x-dx)*gradx.at<double>(y-dy,x-dx) +
                  grady.at<double>(y-dy,x-dx)*grady.at<double>(y-dy,x-dx));
        if (scalarprod/(v1*v2) < parallelity_threshold) {
          angle = atan2(grady.at<double>(y+dy,x+dx),gradx.at<double>(y+dy,x+dx));
          // shift angle bins by pi/n so that zero etc. lie in middle of bin
          if (angle < -pi_n) angle += 2*M_PI;
          i = (int)((angle+pi_n)*nbins/(2*M_PI));
          if (weighted) {
            bin[i] += v1;
            bin[(i+nbins/2)%nbins] += v2;
            sumG += v1+v2;
          } else {
            bin[i]++;
            bin[(i+nbins/2)%nbins]++;
            sumG += 2.0;
          }
        }
      }
    }
    dx = r;
    for (dy=-r+1; dy<r; dy++) {
      scalarprod = gradx.at<double>(y+dy,x+dx)*gradx.at<double>(y-dy,x-dx)
        + grady.at<double>(y+dy,x+dx)*grady.at<double>(y-dy,x-dx);
      if (scalarprod < 0.0) { // only when in opposite directions
        v1 = sqrt(gradx.at<double>(y+dy,x+dx)*gradx.at<double>(y+dy,x+dx) +
                  grady.at<double>(y+dy,x+dx)*grady.at<double>(y+dy,x+dx));
        v2 = sqrt(gradx.at<double>(y-dy,x-dx)*gradx.at<double>(y-dy,x-dx) +
                  grady.at<double>(y-dy,x-dx)*grady.at<double>(y-dy,x-dx));
        if (scalarprod/(v1*v2) < parallelity_threshold) {
          angle = atan2(grady.at<double>(y+dy,x+dx),gradx.at<double>(y+dy,x+dx));
          // shift angle bins by pi/n so that zero etc. lie in middle of bin
          if (angle < -pi_n) angle += 2*M_PI;
          i = (int)((angle+pi_n)*nbins/(2*M_PI));
          if (weighted) {
            bin[i] += v1;
            bin[(i+nbins/2)%nbins] += v2;
            sumG += v1+v2;
          } else {
            bin[i]++;
            bin[(i+nbins/2)%nbins]++;
            sumG += 2.0;
          }
        }
      }
    }
  }

  for (i=0; i<nbins; i++) bin[i] = bin[i]/sumG;
  count = 0;
  for (i=0; i<nbins; i++)
    if (bin[i] > count_threshold) count++;

  max_frequency = *std::max_element(bin.begin(), bin.end());
  result = count/((double)nbins);
  return result;
}
