#ifndef _SYMMETRY_CD_H
#define _SYMMETRY_CD_H

//
// Author:  Christoph Dalitz, Regina Pohle-Froehlich
// Date:    2013/01/03
//

#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include "symmetry_point.h"

//-----------------------------------------------------------------
// Convert float image to greyscale image.
// Arguments:
//  fimg = float image (input)
//  gimg = greyscale image (output)
//-----------------------------------------------------------------
void float2grey(const cv::Mat &fimg, cv::Mat &gimg);

//-----------------------------------------------------------------
// Convert float image to BGR image.
// Arguments:
//  fimg = float image (input)
//  cimg = color image (output)
//-----------------------------------------------------------------
void float2bgr(const cv::Mat &fimg, cv::Mat &cimg);

//-----------------------------------------------------------------
// Gradient normalization function
// Derive your custom normalization routine for special normalizations
// The call operator has two arguments:
//  gx = x-component of gradient (input and output)
//  gy = y-component of gradient (input and output)
//-----------------------------------------------------------------
struct GradientNormalization {
  GradientNormalization(void) {};
  void operator()(double &gx, double &gy) const;
};

//-----------------------------------------------------------------
// Compute symmetry transform from a greyscale image.
// To each point, a symmetry score value and a radius is assigned
// Arguments:
//  grey     = greyscale image (input)
//  result   = symmetry scores as CV_64F image (output)
//  result_r = symmetry radius as CV_64F image (output)
//  max_r    = maximum radius examined (input)
//  gradx    = x-component of gradient (output)
//  grady    = y-component of gradient (output)
//  gradnorm = optional gradient normalization function (default = NULL)
//-----------------------------------------------------------------
void symmetry_cd(const cv::Mat &grey, cv::Mat &result, cv::Mat &result_r, int radius, GradientNormalization *gradnorm=NULL);

//-----------------------------------------------------------------
// Find candidate symmetry centers in symmetry image as local
// maxima of the symmetry score. Returns the number of points found.
// Arguments:
//  symmetry        = symmetry transform image filename (input)
//  candidatepoints = symmetry candidate points (output)
//  windowsize      = window size for local maximum search (input)
//-----------------------------------------------------------------
int points_local_maxima(const cv::Mat &symmetry, SymmetryPointVector &candidatepoints, int windowsize = 3);


//-----------------------------------------------------------------
// Compute the edge dispersion at the given point.
// Arguments:
//  gradx    = x-component of gradient image (input)
//  grady    = y-component of gradient image (input)
//  point    = the point to be examined (input)
//  radius   = the radius of the window to be examined (input)
//  cutoff   = ignore an image border of width cutoff (input)
//-----------------------------------------------------------------
double edge_dispersion_cd(const cv::Mat &gradx, const cv::Mat &grady, const cv::Point &point, int radius = 3, int cutoff = 0);

//-----------------------------------------------------------------
// Compute the edge directedness as the frequency of the most frequent
// gradient direction.
// Arguments:
//  gradx    = x-component of gradient image (input)
//  grady    = y-component of gradient image (input)
//  point    = the point to be examined (input)
//  radius   = the radius of the window to be examined (input)
//  nbins    = number of bins in direction histogram (input)
//  cutoff   = ignore an image border of width cutoff (input)
//-----------------------------------------------------------------
double edge_directedness_cd(const cv::Mat &gradx, const cv::Mat &grady, const cv::Point &point, int radius = 3, int nbins = 16, int cutoff = 0);


//-----------------------------------------------------------------
// Compute the ratio of the the two eigenvalues of the covariance matrix
// of the window with the given radius around point.
// Arguments:
//  image     = complete image (input)
//  point     = the point to be examined (input)
//  radius    = the radius of the window to be examined (input)
//  cutoff   = ignore an image border of width cutoff (input)
//-----------------------------------------------------------------
double coveigenratio_cd(const cv::Mat &image, const cv::Point &point, int radius = 3, int cutoff = 0);

//-----------------------------------------------------------------
// Returns the number of antiparallel gradient bins that have a
// frequency count larger than count_threshold
// inside the window with the given radius around point.
// Arguments:
//  gradx    = x-component of gradient image (input)
//  grady    = y-component of gradient image (input)
//  point    = the point to be examined (input)
//  radius   = the radius of the window to be examined (input)
//  max_frequency = frequency count of largest bin (output)
//  weighted = whether edges shall be weighted by gradient strength (input)
//  count_threshold = threshold when a bin is considered "occupied" (input)
//  nbins    = number of bins in direction histogram (input)
//-----------------------------------------------------------------
double antiparallelity_rpf(const cv::Mat &gradx, const cv::Mat &grady, const cv::Point &point, int radius, double &max_frequency, bool weighted = false, double count_threshold = 0.1, int nbins = 8);


#endif
